/*
 * Kibitz - Query operations
 *
 * (c) 2005-2012 Tomas Valla <tom@ucw.cz>
 *
 * This software may be freely distributed and used according to the terms
 * of the GNU Lesser General Public License. See COPYING for details.
 */

#undef LOCAL_DEBUG

#include <kibitz/lib.h>
#include <ucw/mempool.h>
#include <kibitz/internal.h>
#include <alloca.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

struct kibitz_query *kibitz_new_query(void)
{
	struct mempool *mp = mp_new(4096);
	struct kibitz_query *q = mp_alloc_zero(mp, sizeof(*q));

	q->pool = mp;
	q->max_results = config.query_max_results;
	return q;
}

void kibitz_free_query(struct kibitz_query *q)
{
	mp_delete(q->pool);
}

void kibitz_dump_query(struct kibitz_query *q)
{
	DUMP("Query %p:\n", q);
	DUMP("error_code=%d\n", q->error_code);

	struct kibitz_expr *e;
	CLIST_WALK(e, q->expr)
		DUMP("  expr: word='%s' wlen=%d bonus_factor=%f\n", e->word, e->wlen, e->bonus_factor);

	DUMP("num_results=%d max_results=%d\n", q->num_results, q->max_results);
	struct kibitz_result *r;
	uns i = 0;
	CLIST_WALK(r, q->results) {
		DUMP("  result %d: quality=%d relevance=%d text='%s'\n", i, r->quality,
			r->phrase.relevance, r->phrase.text);
		i++;
	}
}

int kibitz_add_expr(struct kibitz_query *q, byte *word, uns wlen, double bonus_factor)
{
	struct kibitz_expr *e = mp_alloc_zero(q->pool, sizeof(*e));
	e->word = mp_alloc_zero(q->pool, wlen+1);
	memcpy(e->word, word, wlen);
	e->word[wlen] = 0;
	e->wlen = wlen;
	e->bonus_factor = bonus_factor;
	q->num_exprs++;
	clist_add_tail(&q->expr, &e->node);

	return q->error_code;
}

int kibitz_init_query(struct kibitz_query *q)
{
	clist_init(&q->expr);
	clist_init(&q->results);

	return q->error_code;
}

struct trail {
	struct result *results;
	uns num_results;
};

static struct result *prbuf;
#define ASORT_PREFIX(x) phrase_##x
#define ASORT_KEY_TYPE struct result
#define ASORT_ELT(i) prbuf[i]
#define ASORT_LT(x,y) ((x).phrase < (y).phrase)
#include <ucw/sorter/array-simple.h>

#define ASORT_PREFIX(x) quality_##x
#define ASORT_KEY_TYPE struct result
#define ASORT_ELT(i) prbuf[i]
#define ASORT_LT(x,y) ((x).quality > (y).quality)
#include <ucw/sorter/array-simple.h>

static void merge_results(struct kibitz_query *q, struct trail t[], uns tc, uns totalr)
{
	struct trail *te = t+tc;

	ASSERT(totalr);
	prbuf = alloca(totalr * sizeof(*prbuf));

	uns i = 0, j = 0;
	while (t < te) {
		if (j == t->num_results) {
			j = 0;
			t++;
			continue;
		}
		ASSERT(i < totalr);
		prbuf[i++] = t->results[j++];
	}
	ASSERT(i == totalr);

	phrase_sort(i);

	int totq = 0;
	for (j=0, i=0; i<totalr; i++) {
		totq += prbuf[i].quality;
		if (i<totalr-1 && prbuf[i].phrase == prbuf[i+1].phrase)
			continue;
		prbuf[j].phrase = prbuf[i].phrase;
		prbuf[j].quality = totq;
		j++;
		totq = 0;
	}

	quality_sort(j);

	for (i=0; i<MIN(j, q->max_results); i++) {
		struct phrase *p = prbuf[i].phrase;
		struct kibitz_result *r = mp_alloc_zero(q->pool, sizeof(*r)+p->text_len);
		r->quality = prbuf[i].quality;
		memcpy(&r->phrase, p, sizeof(*p) + p->text_len);
		clist_add_tail(&q->results, &r->node);
		q->num_results++;
	}
}

int kibitz_get_results(struct kibitz_query *q)
{
	struct trail trail[q->num_exprs];
	uns tc = 0;

	DBG("Executing search for query %p", q);

	bzero(trail, sizeof(trail));

	struct kibitz_expr *e;
	struct result *r = mp_alloc(q->pool, q->max_results*sizeof(*r));
	uns totalr = 0;
	CLIST_WALK(e, q->expr) {
		uns rc = search_index(e->word, e->wlen, r, q->max_results);
		DBG("search_index for '%s' returned %d results", e->word, rc);
		if (rc) {
			trail[tc].results = r;
			trail[tc].num_results = rc;
			for (uns i=0; i<rc; i++)
				r[i].quality *= e->bonus_factor;
			tc++;
			r = mp_alloc(q->pool, q->max_results*sizeof(*r));
			totalr += rc;
		}
	}

	if (!tc)
		return q->error_code;

	if (tc == 1) {
		for (uns i=0; i<trail->num_results; i++) {
			struct phrase *p = trail->results[i].phrase;
			struct kibitz_result *r = mp_alloc_zero(q->pool, sizeof(*r)+p->text_len);
			r->quality = trail->results[i].quality;
			memcpy(&r->phrase, p, sizeof(*p) + p->text_len);
			clist_add_tail(&q->results, &r->node);
			q->num_results++;
		}
		q->num_results = trail->num_results;
		return q->error_code;
	}
	merge_results(q, trail, tc, totalr);
	return q->error_code;
}
