#!/bin/bash
#
#  Kibitz Search Server Control Script
#  (c) 2002--2003 Martin Mares <mj@ucw.cz>
#  (c) 2006 Vladimir Jelen <vladimir.jelen@netcentrum.cz>
#
# This software may be freely distributed and used according to the terms
# of the GNU Lesser General Public License. See COPYING for details.


set -e

if [ $UID = 0 ] ; then
	cd ${KIBITZ_HOME:-~kibitz/kibitz/run}
	exec su ${KIBITZ_USER:-kibitz} -c "exec bin/kibitz-control $@"
else
	if [ -n "$KIBITZ_HOME" ] ; then cd $KIBITZ_HOME ; fi
fi
if ! [ -f bin/khttpd -a -f etc/kibitz ] ; then
	echo >&2 "kibitz-control: Cannot find Kibitz runtime directory"
	exit 1
fi

eval `bin/config HTTPD Port=3333 StatusFile '*'`
eval `bin/config KibitzKeeper TestRetry=10 TestWait=3 CrashMail RotateLogs CrashWaitThreshold CrashWaitCeiling DaemonPIDFile KeeperPIDFile @TestQuery SwapLock SwapLockTimeout=10`

function log
{
	bin/logger kibitz-control I "$1"
}

function errlog
{
	bin/logger kibitz-control E "$1"
}

if [ -n "$CF_CrashWaitThreshold" ] ; then
	DHO="--pid-file=$CF_KeeperPIDFile"
	CMD=bin/kibitz-keeper
	MSG="khttpd with kibitz-keeper"
else
	DHO="--pid-file=$CF_DaemonPIDFile"
	if [ -n "$CF_StatusFile" ] ; then
		DHO="$DHO --status-file=$CF_StatusFile"
	fi
	CMD=bin/khttpd
	MSG="khttpd"
fi

case "$1" in
	start)		echo -n "Starting $MSG..."
			shift
			# Bash is too weak for writing raceless startup/shutdown scripts,
			# so please forgive us a C helper.
			if bin/daemon-helper --check $DHO ; then
				echo " ALREADY RUNNING."
				exit 0
			fi
			bin/daemon-helper --start $DHO -- $CMD "$@" &
			PID=$!
			RETRY=$CF_TestRetry
			echo -n "."
			while true ; do
				if [ -n "$CF_StatusFile" -a ! -f "$CF_StatusFile" ] ; then
					# khttpd isn't initialized yet, so no point in trying queries
					true
				elif wget --no-proxy -qO /dev/null "http://localhost:$CF_Port/?q=$QUERY&ie=iso-8859-2&oe=iso-8859-2" 2>/dev/null ; then
					# khttpd looks initialized, try sending a trivial query
					echo " done."
					exit 0
				fi
				if ! kill -0 $PID 2>/dev/null ; then
					# The daemon or daemon-helper has exited. Unfortunately, bash doesn't
					# give us any way how to get the return code, so in the rare case
					# khttpd was already running and we haven't spotted it by --check,
					# we would report a failure incorrectly. Fortunately enough, if this
					# happens, the query check above exits with OK status first.
					echo " FAILED."
					exit 1
				fi
				if [ $RETRY = 0 ] ; then
					# The daemon got stuck, so go kill it
					echo " TIMED OUT."
					bin/kibitz-control stop
					exit 1
				fi
				RETRY=$(($RETRY-1))
				echo -n "."
				sleep $CF_TestWait
			done
			;;
	stop)		echo -n "Stopping $MSG... "
			if bin/daemon-helper --stop $DHO ; then
				if [ -n "$CF_CrashWaitThreshold" ] ; then
					# If we use the kibitz-keeper, kill kibitz-keeper first and then kill
					# the orphaned khttpd, if there's any.
					bin/daemon-helper --stop --pid-file=$CF_DaemonPIDFile || true
				fi
				if [ -n "$CF_StatusFile" -a -f "$CF_StatusFile" ] ; then
					rm -f $CF_StatusFile
				fi
				echo "done."
			elif [ $? == 102 ] ; then
				echo "not running."
			else
				echo "FAILED."
				exit 1
			fi
			;;
	restart|reload)	bin/kibitz-control stop
			bin/kibitz-control start
			;;
	cron)		if [ -n "$CF_RotateLogs" ] ; then eval $CF_RotateLogs ; fi
			;;
	test)		[ -z "$CF_TestQuery" ] && exit 0
			echo -n "Testing khttpd..."
			for test in "${CF_TestQuery[@]}" ; do
				echo -n "."
				MINREPLIES=`echo $test | cut -d ' ' -f 1`
				QUERY=`echo $test | cut -d ' ' -f 2-`
				T=tmp/test-$$
				if ! wget --no-proxy -qO $T "http://localhost:$CF_Port/?q=$QUERY&ie=iso-8859-2&oe=iso-8859-2" 2>/dev/null ; then
					rm $T
					echo "FAILED."
					errlog "Query '$QUERY' failed."
					exit 1
				fi
				c=`grep -c "^\[" $T || true`
				rm $T
				if [ $c -lt $MINREPLIES ] ; then
					echo "FAILED."
					errlog "Query '$QUERY' produced only $c replies ($MINREPLIES required)."
					exit 1
				fi
			done
			echo " done."
			;;
	swap)		if [ ! -f "index.new" ] ; then
				log "Nothing to swap"
				exit 0
			fi
			echo -n "Acquiring swap lock..."
			RETRY=$CF_SwapLockTimeout
			while ! mkdir $CF_SwapLock 2>/dev/null ; do
				if [ $RETRY = 0 ] ; then
					echo " TIMED OUT."
					exit 1
				fi
				RETRY=$(($RETRY-1))
				echo -n "."
				sleep 1
			done
			echo " done."
			rm -f "index.old"
			if [ -f "index.idx" ] ; then
				mv "index.idx" "index.old"
			else
				rm -f "index.idx"
			fi
			log "Swapping index"
			mv "index.new" "index.idx"
			if bin/daemon-helper --check $DHO ; then
				bin/daemon-helper --reload --pid-file=$CF_DaemonPIDFile
			else
				bin/kibitz-control start
			fi
			if bin/kibitz-control test ; then
				rmdir $CF_SwapLock
				exit 0
			fi
			if [ -f "index.old" ] ; then
				log "Falling back to old index"
				mv "index.idx" "index.new"
				mv "index.old" "index.idx"
				bin/daemon-helper --reload --pid-file=$CF_DaemonPIDFile
			else
				errlog "No old index to fall back to"
			fi
			if bin/kibitz-control test ; then
				log "Fallback succeeded"
			else
				errlog "Fallback failed"
			fi
			rmdir $CF_SwapLock
			exit 1
			;;
	*)		echo >&2 "Usage: [KIBITZ_USER=<user>] [KIBITZ_HOME=<homedir>] kibitz-control (start|stop|restart|cron|test|swap)"
			exit 1
			;;
esac
